# -*- coding: utf-8 -*-
#
# Univention Directory Listener Module classstructure
#  listener module: classstructure
#
# Copyright 2016 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''		# workaround for PEP 366

import os
import re
import shutil
import traceback

import listener
import univention.debug as ud
import univention.uldap

name        = 'lmz-class-structure-homes'
# TODO prepare to also receive users .. or maybe not??
description = 'Generate class directories and build a link structure to the homes of pupils'
filter      = "(objectClass=univentionGroup)"
attributes  = ['cn', 'uniqueMember', 'homeDirectory', 'univentionObjectType']

log_error   = lambda msg: ud.debug(ud.LISTENER, ud.ERROR, '%s: %s' % (name, msg))
log_warn    = lambda msg: ud.debug(ud.LISTENER, ud.WARN, '%s: %s' % (name, msg))
log_process = lambda msg: ud.debug(ud.LISTENER, ud.PROCESS, '%s: %s' % (name, msg))
log_info    = lambda msg: ud.debug(ud.LISTENER, ud.INFO, '%s: %s' % (name, msg))

class_pattern = re.compile('cn=%s,cn=groups,ou=[^,]+,(ou=[^,]+,)?%s$' % (listener.configRegistry.get('ucsschool/ldap/default/container/pupils', 'schueler'), listener.configRegistry['ldap/base']), re.I)
homedir = listener.configRegistry.get('lmz/classstructure/homedir', '/home')
classbasedir = os.path.join(homedir, listener.configRegistry.get('lmz/classstructure/classbasedir', '_klassen'))
strip_classprefix = listener.configRegistry.get('lmz/classstructure/strip_classprefix', '')


def getmachinesecret():
	""" Reads the machine secret from /etc/machine.secret

	:returns: Returns the machine secret or None if not accessable
	"""
	try:
		listener.setuid(0)
		f = file('/etc/machine.secret')
		pw = f.read().strip('\n')
		f.close()
		return pw
	except Exception, e:
		log_error('error encountered\n%s' % traceback.format_exc())
	finally:
		listener.unsetuid()


_LDAP = None
def init_ldapconnection(force=False):
	""" Initialize LDAP connection

	:returns: ldap connection object
	"""
	global _LDAP
	if _LDAP is None or force:
		port = listener.configRegistry.get('ldap/port')
		if not port:
			port = listener.configRegistry.get('ldap/server/port')
		port = int(port)
		args = {'base' : listener.configRegistry.get('ldap/base'),
				'port' : port,
				'host' : listener.configRegistry.get('ldap/server/name')
				}
		pw = getmachinesecret()
		if pw is not None:
			# perform an authenticated bind
			args['binddn'] = listener.configRegistry.get('ldap/hostdn')
			args['bindpw'] = pw
		_LDAP = univention.uldap.access(**args)
	return _LDAP


def ldapget(dn, attrlist=None):
	""" Retrieve a single LDAP object

	:dn: DN of the object that shall be retrieved
	:attrlist: List of attributes that shall be retrieved
	:returns: Return the ldap object. The Result is a dict of the form
	          {attr1: (value1, value2), ..}

	Example:
	# this is safe because even if ldapget doesn't find anything, it still
	# returns an empty dict ({})
	for k, v in ldapget('uid=jebersba,...').iteritems():
		print k
	"""
	ldap = init_ldapconnection()
	assert ldap is not None
	res = ldap.get(dn=dn, attr=attrlist)
	if not res:
		# workaround in case LDAP server is not reachable, retrieve LDAP base
		if not ldap.get(dn=listener.configRegistry['ldap/base'], attr=['objectClass']):
			log_warn('forcing LDAP reconnect')
			init_ldapconnection(force=True)
			res = ldap.get(dn=dn, attr=attrlist)
			if not res and not ldap.get(dn=listener.configRegistry['ldap/base'], attr=['objectClass']):
				log_error('failed to reconnect to LDAP server')
	return res


def retrieve_userhome(dn):
	""" Retrieve user's home directory for the given dn

	:dn: user's DN
	:returns: home directory or None if user isn't found

	"""
	obj = ldapget(dn, ('homeDirectory', ))
	if obj and obj['homeDirectory']:
		return obj['homeDirectory'][0]


def check_classbasedir():
	""" Check if classbasedir exists
	:returns: Return True if classbasedir exists

	"""
	if os.path.exists(classbasedir):
		if os.path.isdir(classbasedir):
			return True
	return False


def check_classdir(classdir):
	""" Check if classdir exists
	:returns: Return True if classdir exists

	"""
	if os.path.exists(classdir):
		if os.path.isdir(classdir):
			return True
	return False


def delete_symlink(classdir, userhomedir):
	""" Delete a symlink in class dir that points to the user's home directory

	:classdir: Directory of the specific class
	:userhomedir: User's home directory
	:returns: None
	"""
	log_info('deleting symlink to user\'s home directory %s in %s' % (classdir, userhomedir))
	listener.setuid(0)
	try:
		link = os.path.sep.join((classdir, os.path.basename(userhomedir)))
		if os.path.islink(link):
			os.remove(link)
	except Exception, e:
		log_error('error encountered\n%s' % traceback.format_exc())
	finally:
		listener.unsetuid()


def create_symlink(classdir, userhomedir):
	""" Create a symlink in class dir to the user's home directory

	:classdir: Directory of the specific class
	:userhomedir: User's home directory
	:returns: None
	"""
	log_info('creating symlink to user\'s home directory %s in %s' % (classdir, userhomedir))
	listener.setuid(0)
	try:
		os.symlink(userhomedir, os.path.sep.join((classdir, os.path.basename(userhomedir))))
	except Exception, e:
		log_error('error encountered\n%s' % traceback.format_exc())
	finally:
		listener.unsetuid()


def build_classdir_path(classobj):
	""" Build the class dir's path from the class object

	:classobj: ldap class object
	:returns: class dir path
	"""
	classname = classobj['cn'][0]
	if classname.startswith(strip_classprefix):
		classname = classname[len(strip_classprefix):]
	return os.path.sep.join((classbasedir, classname))


def delete_classdir(classobj):
	""" Delete a class dir

	:classobj: ldap class object
	:returns: True if the dir has been deleted
	"""
	classdir = build_classdir_path(classobj)
	log_info('deleting class dir %s' % classdir)
	if os.path.exists(classdir):
		if os.path.isdir(classdir):
			listener.setuid(0)
			try:
				shutil.rmtree(classdir)
			except Exception, e:
				log_error('error encountered\n%s' % (traceback.format_exc()))
			finally:
				listener.unsetuid()
			return True
		else:
			log_error('class dir exists but is no directory %s' % classdir)
	else:
		log_warn('class dir doesn\'t exist anymore %s' % classdir)
	return False


def process_uniqueMembers(members, classdir, func):
	""" Process uniqueMemebers with func

	:members: List of uniqueMember dns
	:func: either create_symlink or delete_symlink function pointer
	:returns: None

	"""
	for uniqueMember in members:
		userhomedir = retrieve_userhome(uniqueMember)
		if userhomedir:
			if userhomedir.startswith(homedir):
				func(classdir, userhomedir)
			else:
				log_info('ignoring user home (%s) outside homedir %s' % (userhomedir, homedir))
		else:
			log_info('ignoring non-existent user %s' % (uniqueMember, ))


def create_classdir(classobj):
	""" Create class dir with symlinks for pupils' homes

	:classobj: ldap class object
	:returns: None
	"""
	classdir = build_classdir_path(classobj)
	log_info('creating class dir %s' % classdir)
	listener.setuid(0)
	try:
		os.makedirs(classdir, mode=0755)
	except Exception, e:
		log_error('error encountered\n%s' % traceback.format_exc())
	finally:
		listener.unsetuid()

	# create symlinks
	process_uniqueMembers(classobj.get('uniqueMember', []), classdir, create_symlink)


def modify_classdir(new_classobj, old_classobj):
	""" Modify class dir with symlinks for pupils' homes

	:new_classobj: ldap class object
	:old_classobj: ldap class object
	:returns: None
	"""
	classdir = build_classdir_path(new_classobj)
	log_info('modifying class dir %s' % classdir)
	if not check_classdir(classdir):
		log_error('class dir doesn\'t exist %s' % classdir)
		return False

	new_members = frozenset(new_classobj.get('uniqueMember', []))
	old_members = frozenset(old_classobj.get('uniqueMember', []))

	# delete/create symlinks
	process_uniqueMembers(old_members.difference(new_members), classdir, delete_symlink)
	process_uniqueMembers(new_members.difference(old_members), classdir, create_symlink)

def initialize():
	# initialize base directory
	if os.path.exists(classbasedir):
		if not os.path.isdir(classbasedir):
			log_error('exists but is not a directory' % classbasedir)
		else:
			log_warn('reinitializing module, deleting %s' % classbasedir)
			listener.setuid(0)
			try:
				shutil.rmtree(classbasedir)
			except Exception, e:
				log_error('error encountered\n%s' % traceback.format_exc())
			finally:
				listener.unsetuid()

	log_info('creating base dir %s' % classbasedir)
	listener.setuid(0)
	try:
		os.makedirs(classbasedir, mode=0755)
	except Exception, e:
		log_error('error encountered\n%s' % traceback.format_exc())
	finally:
		listener.unsetuid()


def handler(dn, new, old):
	log_process('processing %s' % dn)
	if not check_classbasedir():
		log_error('base dir doesn\'t exist %s' % classbasedir)
		return False
	class_match = class_pattern.search(dn)
	if not class_match:
		log_info('ignoring non-class group %s' % dn)
		return

	# add class
	if new and not old:
		create_classdir(new)
	# modify class
	elif new and old:
		modify_classdir(new, old)
	# delete class
	else:
		delete_classdir(old)

# vi: ft=python:tw=80:sw=4:ts=4:sts=4:noet
